/*
 * Copyright (c) 2018      Mellanox Technologies, Inc.
 *                         All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef GDS_DS12_FILE_H
#define GDS_DS12_FILE_H

#include <src/include/pmix_config.h>
#include <pmix_common.h>

extern pmix_common_dstore_file_cbs_t pmix_ds12_file_module;
extern pmix_common_dstore_file_cbs_t pmix_ds20_file_module;

#endif // GDS_DS12_FILE_H
